/* Copyright 2000,2001 John Fremlin
   Distributed under GPL, see file COPYING
*/

#define _GNU_SOURCE

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <termio.h>
#include <signal.h>
#include <err.h>
#include "config.h"
#include "system.h"

#ifndef USE_TERMINAL_PIDS
#define USE_TERMINAL_PIDS 1
#endif

static inline
void
pids_dump(void)
{
  pid_t pid = getpid();

  warnx(_("dumping (%s %s) pid information"),PACKAGE,VERSION);

  if(!isatty(STDIN_FILENO))
    warnx(_("warning: fd %d is not a tty"),STDIN_FILENO);


#if USE_TERMINAL_PIDS
  warnx(_("%d is terminal controlling process group"),
	  tcgetpgrp(0));
  warnx(_("%d is terminal controlling session"),
	  tcgetsid(0));
#endif

  warnx(_("%d is process group"),
	  getpgrp());
  warnx(_("%d is session"),
	  getsid(pid));
  warnx(_("%d is process"),
	  pid);

}
#if 0
void own_tty(int fd) 
{
  pid_t pid=getpid();
  
  if(ioctl(fd,TIOCSCTTY,pid))
      warn(_("warning: can't become tty session controller"));
  
  /* can't change pgid of terminal unless you're in the right session
     group, so this usually fails unless TIOSCSTTY succeeded first */
  if(tcsetpgrp(fd, pid))
    warn(_("warning: can't tcsetpgrp"));
}

void
reset_tty(int fd)
{
#ifdef HAVE_VHANGUP
  struct sigaction old, off;
  sigemptyset (&off.sa_mask);
  off.sa_flags = 0;
  off.sa_handler = SIG_IGN;
  if(sigaction (SIGHUP, &off,&old)<0){
    perror("sigaction");
    return;
  }

  if(vhangup()<0) /* wastes all fds pointing to tty */
    {
      perror("vhangup");
      close(fd);
    }
  
  if(sigaction(SIGHUP,&old,0)<0){
    perror("restoring sigaction");
    return;
  }
#else
  close(fd);
#endif
}
#endif

static
int
open_term(const char*device)
{

  /* Test we can open the tty. Then we close all fds and open again,
     so that it becomes our controlling terminal even if we already
     had one */
  int fd = open(device,O_RDWR|O_NONBLOCK,0);
  int i;
  
  if(fd==-1)
    warnx(_("problem opening tty %s"),device);

  if(!isatty(fd))
    warnx(_("warning: %s is not a tty"),device);

  for(i=0;i<=fd;i++)
    close(i);

  fd=open(device,O_RDWR,0);
  if(fd==-1)abort();
  
  for(i=0;i<3;i++)
    if(fd!=i)if(dup(fd)!=i)abort();

  if(fd>2)close(fd);
  
  return 0;
}

static
int
pids_admin(void)
{
  pid_t pid=getpid();
  pid_t session=getsid(pid);
  pid_t group=getpgid(pid);
  
  if(pid==session)/*someone (init?) has kindly given us our own session*/
    {
      if(group!=pid)
	setpgid(pid,pid);/*but not our own process group*/
    }
  else
    {
      /* We can't setsid successfully if we already are leaders of a
	 process group (process id == process group id), so we fork for a
	 new pid (which will end up in the current process group) */

      if(group==pid)
	{
	  pid_t child=fork();
	  if(child)
	    {
	      /* in parent: to wait or not to wait? init would go
		 respawn mad if we don't, but init is not like a
		 shell; it gives us our own session. If we do wait we
		 waste a process entry. */
	      exit(0);
	    }
	  
	  pid=getpid();
	  /* in child */
	  if(setsid()==-1)
	    /* note that setsid gives us our own process group as well */
	    {
	      warn(_("warning: can't setsid"));
	      pids_dump();
	    }
	}
    }
  
#if 0  
  /* On vhangup called from login, we get killed (login blocks the
     signal as below, so it's ok if we've exec'd login) */
  signal(SIGHUP,SIG_IGN);
#endif
  
  return 0;
}  

int
main(int argc,char*argv[],char*env[])
{
  gettext_init();
  
  if(argc<2)
    {
      warnx(_("too few arguments"));
      printf(_("%s %s\t\t(c) 2000,2001 John Fremlin\n"),
	     PACKAGE,VERSION);
      printf(_("syntax: [terminal device] [optional program to run]\n\n"));
      printf(_("If no other program is specified, %s is run.\n"),
	     DEFAULT_AGENT);
      return 1;
    }
  
  if(pids_admin())
    return 2;

  if(open_term(argv[1]))
    return 3;
  
  if(argv[2])
    {
      if(execvp(argv[2],argv+2))
	{
	  warn(_("problem execvping %s"),argv[2]);
	  return 4;
	}
    }
  else
    {
      if(execl(DEFAULT_AGENT,DEFAULT_AGENT,0))
	{
	   warnx(_("problem execling login agent %s"),DEFAULT_AGENT);
	   return 5;
	}
    }
  
  return 0;
}

